function [info] = CoalitionLinkage(A,net,params,mtd)
% info = CoalitionLinkage(A,net,params,mtd)
% 
% perform Coalition Formation using a Linkage algorithm
% <A> is the set of identifiers of the agents
% <net> network of connections: net_ij = 1 --> i and j can be in the same
%   coalition
% <params> is a cell array containing the parameters used to call
%   the function CoalVal
% <mtd> defines the method used to join the coalitions
%   - 'g': Gain Link (Default)
%   - 's': Single Link
%   - 'c': Complete Link
%   - 'a': Average Link
%
%
% 
% the output is a structure <info> which contains, among others
% 
% info.C                --> Best Coalition Structure found (CS~)
% info.OptVal           --> value of CS~
% info.AvgCoalSize      --> average size of the coalitions in CS~
% info.MinCoalSize      --> min size of the coalitions in CS~
% info.MaxCoalSize      --> max size of the coalitions in CS~
% info.NCoal            --> number of coalitions in CS~
% info.elapsedTime      --> elapsed time
% info.SingleCoalVal    --> v(S) = sum_{a \in A} v(a)
% info.vA               --> value of the Grand Coalition v(A) 
% info.Hval         


if nargin < 4
    mtd = 'g'; % default GainLink
end

% make sure A is a row
A = reshape(A,[1,length(A)]);

% step 1: initialize: every user is a cluster
tic;
nel = length(A);
for i = 1:nel
    C{i} = i;
    v(i) = CoalVal(A(C{i}),params);
end
% initialize the matrix of Values (the matrix of similarity)
W = -inf(nel);
for i = 1:nel
    for j = i+1:nel
        if net(i,j) == 1
            W(i,j) = CoalVal(A([C{i} C{j}]),params) - v(i) - v(j);
        end
    end
end
Worig = W;

% repeat:
% 1. find maximum value
% 2. join the corresponding two coalition
% 3. update v, W and net
% until the maximum value is positive or we have only one cluster
t = 0;
maxvalue = inf;
while (maxvalue>=-0.0000001 && length(C) ~=1)
    t = t+1;
    fprintf('Iteration %d\n',t);
    Hierarchy{t} = C; % store current clustering
    % 1. find maximum value
    [x,i]=max(W);
    [maxvalue,i2]=max(x);
    if maxvalue >=-0.0000001
        tojoin=[i(i2),i2];
        Hval(t) = maxvalue;
        
         % 2. join the corresponding two coalition
        C{end+1} = [C{tojoin(1)} C{tojoin(2)}];
        C(tojoin) = [];
        nclus = length(C);
       
        % update net
        net(end+1,:) = 0; % new row
        net(:,end+1) = 0; % new column
        net(end,end)=1; % the new cluster can be connected with itself
        for i=1:nclus+1 % we will remove lines after
            if net(tojoin(1),i) + net(tojoin(2),i) >= 1
                % connect if  cluster to be joined are connected
                net(end,i) = 1;
                net(i,end) = 1;
            end
        end
        net(tojoin,:) = [];
        net(:,tojoin) = [];
        % update W
        
        % remove rows and columns from W
        W(tojoin,:) = [];
        W(:,tojoin) = [];
        % add the last row
        W(end+1,:) = -inf; % W is upper triangle
        W(end,end+1) = -inf; % add a column
        
        % add the last column/row
        % here it depends on the method
        valueCend = CoalVal(A([C{end}]),params);
        for i=1:nclus-1
            if net(i,end) == 1
                Wtemp = Worig([C{i}],[C{end}]);
                switch mtd(1)
                    case 'g' % GainLinkage
                        W(i,end) = CoalVal(A([C{i} C{end}]),params) - CoalVal(A([C{i}]),params) - valueCend;
                    case 's' % single link
                        % min distance --> max similarity
                        W(i,end) = max(Wtemp(:));
                    case 'c' % complete link
                        % max distance --> min similarity
                        W(i,end) = min(Wtemp(:));
                    case 'a' % average link
                        W(i,end) = mean(Wtemp(:));
                end
            else
                W(i,end) = -inf;
            end
            W(end,i) = W(i,end);
        end
    end
end

% statistics and output
nclus = length(C);
nA = length(A);
vc = zeros(1,nclus);
siz = zeros(1,nclus);
for c=1:nclus
    vc(c) = CoalVal(A([C{c}]),params);
    siz(c) = length(C{c});
end
OptVal = sum(vc);
AvgCoalSize = mean(siz);
MinCoalSize = min(siz);
MaxCoalSize = max(siz);
NCoal = nclus;

va = zeros(1,nA);
for i = 1:nA
    va(i) = CoalVal(A(i),params);
end
SingleCoalVal = sum(va);
elapsedTime = toc;
vA = CoalVal(A,params); % value of the grand coalition

info.OptVal = OptVal;
info.SingleCoalVal = SingleCoalVal;
info.AvgCoalSize = AvgCoalSize;
info.MinCoalSize = MinCoalSize;
info.MaxCoalSize = MaxCoalSize;
info.NCoal = NCoal;
info.elapsedTime = elapsedTime;
info.vA = vA;
info.Hval = Hval;
info.C = C;


